local AddonName, GGL = ...
-- loader for classic things

if not C_BattleNet then
    local GameGUIDCache = {}
    local AccountGUIDCache = {}
    local C_BattleNet = {}

    local function GetFriendInfo(presenceID, accountName, battleTag, isBattleTagPresence, characterName, bnetIDGameAccount, client, isOnline, lastOnline, isAFK, isDND, messageText, noteText, isRIDFriend, messageTime, canSoR, isReferAFriend, canSummonFriend)
        if battleTag then
            return {
                bnetAccountID = presenceID,
                accountName = accountName,
                battleTag = battleTag,
                isFriend = true,
                isBattleTagFriend = isRIDFriend,
                lastOnlineTime = lastOnline,
                isAFK = isAFK,
                isDND = isDND,
                appearOffline = false,
                customMessage = messageText,
                customMessageTime = messageTime,
                note = noteText,
                rafLinkType = isReferAFriend and 2 or 0,
                gameAccountInfo = {},
            }
        end
    end

    function C_BattleNet.GetFriendAccountInfo(friendIndex)
        return GetFriendInfo(BNGetFriendInfo(friendIndex))
    end

    function C_BattleNet.GetAccountInfoByID(accountID)
        if not accountID then return end
        if BNIsSelf(accountID) then
            local presenceID, battleTag, toonID = BNGetInfo()
            GameGUIDCache[UnitGUID("player")] = 99999
            return GetFriendInfo(toonID, nil, battleTag, true, nil, 99999)
        else
            return GetFriendInfo(BNGetFriendInfoByID(accountID))
        end
    end

    C_BattleNet.GetFriendNumGameAccounts = BNGetNumFriendGameAccounts

    function C_BattleNet.GetFriendGameAccountInfo(friendIndex, accountIndex)
        local hasFocus, characterName, client, realmName, realmID, faction, race, class, guild, zoneName, level, gameText, broadcastText, broadcastTime, canSoR, bnetIDGameAccount, presenceID, isGameAFK, isGameBusy, characterGUID, wowProjectID, wowClassicRealmName = BNGetFriendGameAccountInfo(friendIndex, accountIndex)
        if characterGUID and not GameGUIDCache[characterGUID] then
            GameGUIDCache[characterGUID] = bnetIDGameAccount
        end
        if characterName then
            return characterName and {
                gameAccountID = bnetIDGameAccount,
                isOnline = true,
                isGameBusy = isGameBusy,
                isGameAFK = isGameAFK,
                characterName = characterName,
                realmName = realmName,
                realmDisplayName = realmName,
                realmID = realmID,
                factionName = faction,
                raceName = race,
                className = class,
                areaName = zoneName,
                characterLevel = level,
                richPresence = format("%s - %s", characterName or "", realmName or ""),
                playerGuid = characterGUID,
                hasFocus = hasFocus,
                wowProjectID = wowProjectID,
                clientProgram = client
            }
        end
    end

    function C_BattleNet.GetGameAccountInfoByID(bnetIDGameAccount)
        if not bnetIDGameAccount then return end
        local hasFocus, characterName, client, realmName, realmID, faction, race, class, guild, zoneName, level, gameText, broadcastText, broadcastTime, canSoR, bnetIDGameAccount2, presenceID, isGameAFK, isGameBusy, characterGUID, wowProjectID, wowClassicRealmName  = BNGetGameAccountInfo(bnetIDGameAccount)
        if characterGUID and not GameGUIDCache[characterGUID] then
            GameGUIDCache[characterGUID] = bnetIDGameAccount
        end
        if bnetIDGameAccount == 99999 then
            return {
                gameAccountID = 99999,
                isOnline = true,
                isGameBusy = UnitIsDND("player"),
                isGameAFK = UnitIsAFK("player"),
                characterName = UnitName("player"),
                realmName = GetRealmName(),
                realmDisplayName = GetRealmName(),
                realmID = GetRealmID(),
                factionName = UnitFactionGroup("player"),
                raceName = UnitRace("player"),
                className = UnitClassBase("player"),
                areaName = GetRealZoneText(),
                characterLevel = UnitLevel("player"),
                richPresence = format("%s - %s", UnitName("player") or "", GetRealmName() or ""),
                playerGuid = UnitGUID("player"),
                hasFocus = false,
                wowProjectID = _G.WOW_PROJECT_ID,
                clientProgram = _G.BNET_CLIENT_WOW
            }
        elseif characterName then
            return {
                gameAccountID = bnetIDGameAccount,
                isOnline = true,
                isGameBusy = isGameBusy,
                isGameAFK = isGameAFK,
                characterName = characterName,
                realmName = realmName,
                realmDisplayName = realmName,
                realmID = realmID,
                factionName = faction,
                raceName = race,
                className = class,
                areaName = zoneName,
                characterLevel = level,
                richPresence = format("%s - %s", characterName or "", realmName or ""),
                playerGuid = characterGUID,
                hasFocus = hasFocus,
                wowProjectID = wowProjectID,
                clientProgram = client
            }
        end
    end

    local function UpdateAccountGUIDLists()
        local numAccounts = BNGetNumFriends()
        if numAccounts and numAccounts > 0 then
            for i = 1, numAccounts do
                C_BattleNet.GetFriendAccountInfo(i)
                local numGameAccounts = BNGetNumFriendGameAccounts(i)
                for j = 1, numGameAccounts do
                    local p = C_BattleNet.GetFriendGameAccountInfo(i, j)
                    if p and p.playerGuid then
                        AccountGUIDCache[p.playerGuid] = i
                    end
                end
            end
        end
    end

    -- BNGetGameAccountInfoByGUID, BNGetAccountInfoByGUID = C_BattleNet.GetGameAccountInfoByGUID, C_BattleNet.GetAccountInfoByGUID
    function C_BattleNet.GetGameAccountInfoByGUID(guid)
        if not GameGUIDCache[guid] then
            UpdateAccountGUIDLists(guid)
        end
        return C_BattleNet.GetGameAccountInfoByID(GameGUIDCache[guid])
    end

    function C_BattleNet.GetAccountInfoByGUID(guid)
        if not AccountGUIDCache[guid] then
            UpdateAccountGUIDLists(guid, true)
        end
        return C_BattleNet.GetAccountInfoByID(AccountGUIDCache[guid])
    end

    GGL.C_BattleNet = C_BattleNet
end

if not EJ_GetInstanceForMap then
    local INSTANCE_TO_EJ_ID = {
        [33] = 64, -- Shadowfang Keep - 0
        [34] = 238, -- Stormwind Stockade - 0
        [36] = 63, -- Deadmines - 0
        [43] = 240, -- Wailing Caverns - 0
        [47] = 234, -- Razorfen Kraul - 0
        [48] = 228, -- Blackfathom Deeps - 0
        [70] = 239, -- Uldaman - 0
        [90] = 231, -- Gnomeregan - 0
        [109] = 237, -- Sunken Temple - 0
        [129] = 233, -- Razorfen Downs - 0
        [189] = 316, -- Scarlet Monastery - 0 ###
        [209] = 241, -- Zul'Farrak - 0
        [229] = 229, -- Blackrock Spire - 0
        [230] = 228, -- Blackrock Depths - 0
        [249] = 760, -- Onyxia's Lair - 0 ###
        [269] = 255, -- Opening of the Dark Portal - 1
        [289] = 246, -- Scholomance - 0 ###
        [309] = 76, -- Zul'Gurub - 0 ###
        [329] = 236, -- Stratholme - 0
        [349] = 232, -- Maraudon - 0
        [389] = 226, -- Ragefire Chasm - 0
        [409] = 741, -- Molten Core - 0
        [429] = 230, -- Dire Maul - 0
        [469] = 742, -- Blackwing Lair - 0
        [509] = 743, -- Ruins of Ahn'Qiraj - 0
        [531] = 744, -- Ahn'Qiraj Temple - 0
        [532] = 745, -- Karazhan - 1
        [533] = 754, -- Naxxramas - 0 ###
        [534] = 750, -- The Battle for Mount Hyjal - 1
        [540] = 259, -- Hellfire Citadel: The Shattered Halls - 1
        [542] = 256, -- Hellfire Citadel: The Blood Furnace - 1
        [543] = 248, -- Hellfire Citadel: Ramparts - 1
        [544] = 747, -- Magtheridon's Lair - 1
        [545] = 261, -- Coilfang: The Steamvault - 1
        [546] = 262, -- Coilfang: The Underbog - 1
        [547] = 260, -- Coilfang: The Slave Pens - 1
        [548] = 748, -- Coilfang: Serpentshrine Cavern - 1
        [550] = 749, -- Tempest Keep - 1
        [552] = 254, -- Tempest Keep: The Arcatraz - 1
        [553] = 257, -- Tempest Keep: The Botanica - 1
        [554] = 258, -- Tempest Keep: The Mechanar - 1
        [555] = 253, -- Auchindoun: Shadow Labyrinth - 1
        [556] = 259, -- Auchindoun: Sethekk Halls - 1
        [557] = 250, -- Auchindoun: Mana-Tombs - 1
        [558] = 247, -- Auchindoun: Auchenai Crypts - 1
        [560] = 251, -- The Escape From Durnholde - 1
        [564] = 751, -- Black Temple - 1
        [565] = 746, -- Gruul's Lair - 1
        [568] = 77, -- Zul'Aman - 1 ###
        [580] = 752, -- The Sunwell - 1
        [585] = 249, -- Magister's Terrace - 1
        [598] = 752, -- Sunwell Fix (Unused) - 1
    }

    local INSTANCE_TO_MAP_ID = {
        [33] = 64, -- Shadowfang Keep - 0
        [34] = 238, -- Stormwind Stockade - 0
        [36] = 63, -- Deadmines - 0
        [43] = 240, -- Wailing Caverns - 0
        [47] = 234, -- Razorfen Kraul - 0
        [48] = 228, -- Blackfathom Deeps - 0
        [70] = 239, -- Uldaman - 0
        [90] = 231, -- Gnomeregan - 0
        [109] = 237, -- Sunken Temple - 0
        [129] = 233, -- Razorfen Downs - 0
        [189] = 316, -- Scarlet Monastery - 0 ###
        [209] = 241, -- Zul'Farrak - 0
        [229] = 229, -- Blackrock Spire - 0
        [230] = 228, -- Blackrock Depths - 0
        [249] = 760, -- Onyxia's Lair - 0 ###
        [269] = 255, -- Opening of the Dark Portal - 1
        [289] = 246, -- Scholomance - 0 ###
        [309] = 76, -- Zul'Gurub - 0 ###
        [329] = 236, -- Stratholme - 0
        [349] = 232, -- Maraudon - 0
        [389] = 226, -- Ragefire Chasm - 0
        [409] = 741, -- Molten Core - 0
        [429] = 230, -- Dire Maul - 0
        [469] = 742, -- Blackwing Lair - 0
        [509] = 743, -- Ruins of Ahn'Qiraj - 0
        [531] = 744, -- Ahn'Qiraj Temple - 0
        [532] = 745, -- Karazhan - 1
        [533] = 754, -- Naxxramas - 0 ###
        [534] = 750, -- The Battle for Mount Hyjal - 1
        [540] = 259, -- Hellfire Citadel: The Shattered Halls - 1
        [542] = 256, -- Hellfire Citadel: The Blood Furnace - 1
        [543] = 248, -- Hellfire Citadel: Ramparts - 1
        [544] = 747, -- Magtheridon's Lair - 1
        [545] = 261, -- Coilfang: The Steamvault - 1
        [546] = 262, -- Coilfang: The Underbog - 1
        [547] = 260, -- Coilfang: The Slave Pens - 1
        [548] = 748, -- Coilfang: Serpentshrine Cavern - 1
        [550] = 749, -- Tempest Keep - 1
        [552] = 254, -- Tempest Keep: The Arcatraz - 1
        [553] = 257, -- Tempest Keep: The Botanica - 1
        [554] = 258, -- Tempest Keep: The Mechanar - 1
        [555] = 253, -- Auchindoun: Shadow Labyrinth - 1
        [556] = 259, -- Auchindoun: Sethekk Halls - 1
        [557] = 250, -- Auchindoun: Mana-Tombs - 1
        [558] = 247, -- Auchindoun: Auchenai Crypts - 1
        [560] = 251, -- The Escape From Durnholde - 1
        [564] = 751, -- Black Temple - 1
        [565] = 746, -- Gruul's Lair - 1
        [568] = 77, -- Zul'Aman - 1 ###
        [580] = 752, -- The Sunwell - 1
        [585] = 249, -- Magister's Terrace - 1
        [598] = 752, -- Sunwell Fix (Unused) - 1
    }

    function GGL.EJ_GetInstanceForMap(mapID)
        local name, instanceType, difficultyID, difficultyName, maxPlayers, dynamicDifficulty, isDynamic, instanceID = GetInstanceInfo()
        return INSTANCE_TO_EJ_ID[instanceID]
    end

    function GGL.C_Map_GetBestMapForUnit(unit)
        local name, instanceType, difficultyID, difficultyName, maxPlayers, dynamicDifficulty, isDynamic, instanceID = GetInstanceInfo()
        if instanceID and INSTANCE_TO_MAP_ID[instanceID] then
            return 101
        else
            return C_Map.GetBestMapForUnit(unit)
        end
    end
end