local AddonName, GGL = ...

local AceGUI = LibStub("AceGUI-3.0")
local GGL_Templates = GGL.Templates
local L = GGL.Locale

local Statistics = {}
GGL.GUI.Statistics = Statistics

-- lua

-- wow

-- GGL
local GetMoneyText = GGL.GetMoneyText
local GetMoneyTextShort = GGL.GetMoneyTextShort

-- ##################
-- GUI - shared
-- ##################
local ExportFrame

local Templates
do
    Templates = {
        ["ShowExportFrame"] = function(data, exportType)
            if not ExportFrame then
                local frame = AceGUI:Create("Frame")
                frame:SetTitle(L["Export"])
                --frame:SetStatusText(listName)
                frame:SetLayout("Fill")

                local multiEditbox = AceGUI:Create("MultiLineEditBox")
                multiEditbox.frame:SetClipsChildren(true)
                multiEditbox.editBox:SetMaxBytes(nil)
                multiEditbox.editBox:SetScript("OnEscapePressed", function()
                    frame:Close()
                end)
                multiEditbox.editBox:SetScript("OnMouseUp", function()
                    multiEditbox.editBox:HighlightText()
                end)
                multiEditbox.editBox:SetScript("OnChar", function(self)
                    multiEditbox:SetText(self.exportText)
                    multiEditbox.editBox:HighlightText()
                end)
                frame.multiEditbox = multiEditbox
                frame:AddChild(multiEditbox)

                ExportFrame = frame
            end

            local multiEditbox = ExportFrame.multiEditbox

            local exportText
            if exportType == "csv" then
                exportText = GGL.Export:ExportAsCSV(data, ",", true)
            end

            multiEditbox.editBox.exportText = exportText or ""
            multiEditbox:SetText(exportText or "")
            ExportFrame:Show()

            multiEditbox.editBox:HighlightText()
            multiEditbox.editBox:SetFocus()
            return ExportFrame
        end,

        ["GetItemQualityBox"] = function(content)
            local qualityGroup = AceGUI:Create("SimpleGroup")
            qualityGroup:ClearAllPoints()
            qualityGroup.frame:SetParent(content)
            qualityGroup:SetLayout("List")
            qualityGroup.barList = {}

            for quality = (#GGL.ITEM_QUALITY_INFO-1), 0, -1 do
                local qData = GGL.ITEM_QUALITY_INFO[quality]

                local bar = AceGUI:Create("GoldGoblinLootGUIWidget-ProgressBar")
                bar:SetFullWidth(true)
                bar:SetBarColor(qData.color.r, qData.color.g, qData.color.b)
                bar:SetPercent(true,true)

                qualityGroup:AddChild(bar)
                qualityGroup.barList[quality] = bar
            end

            return qualityGroup
        end,

        ["GetProgressBar"] = function(content)
            local barContainer = AceGUI:Create("SimpleGroup")
            barContainer:ClearAllPoints()
            barContainer.frame:SetParent(content)
            barContainer:SetLayout("Fill")
            barContainer.bar = {}

            local bar = AceGUI:Create("GoldGoblinLootGUIWidget-ProgressBar")
            bar:SetFullWidth(true)
            bar:SetBarColor(0, 1, 0)
            bar:SetPercent(true,false)

            barContainer:AddChild(bar)

            barContainer.bar = bar

            return barContainer
        end,

        ["GetBoxWithHeader"] = function(parent)
            local frame = GGL.GUI:CreateBoxFrame(parent)
            frame:SetSize(200,100)

            local text = frame:CreateFontString(nil, "BORDER", "GameFontNormal", 2)
            text:SetFontObject(GGL.GetFont(12, "THICKOUTLINE"))
            text:SetPoint("CENTER", frame, "TOP", 0, -2)
            text:SetJustifyH("CENTER")
            text:SetJustifyV("MIDDLE")
            text:SetHeight(12)
            text:SetWordWrap(false)
            text:SetNonSpaceWrap(false)

            local content = CreateFrame("FRAME", nil, frame)
            content:SetPoint("TOPLEFT", frame, 8, -13)
            content:SetPoint("BOTTOMRIGHT", frame, -8, 4)

            -- header fix 5
            -- +17

            frame.text = text
            frame.content = content

            return frame
        end,
    }
end
GGL.GUI.Statistics.Templates = Templates

-- ##################
-- Base
-- ##################

-- ##################
-- GUI - QuickInfo
-- ##################
local QuickInfoFrame
local TOP_ITEMS_LIST_SIZE = 5
local QIF_BUTTON_FRAME_SIZE = 40
local PLAYER_PLAYING_FORMAT, PLAYER_PLAYING_FORMAT_COLOR = "Playing %s-%s", "Playing |c%s%s|r-%s"
local ITEM_NUM_FORMAT = "%d x %s"
local QUALITY_TEXT_FORMAT = "%d - %d%%"
local PERCENT_NUMBER_FORMAT = "%d%%"

local function CreateNewButton(parent)
    local button = CreateFrame("Frame", nil, parent)
    button:SetSize(QIF_BUTTON_FRAME_SIZE, QIF_BUTTON_FRAME_SIZE)

    local icon = button:CreateTexture(nil, "BACKGROUND")
    icon:SetAllPoints(button)
    button.icon = icon

    local textBG = button:CreateTexture(nil, "BACKGROUND")
    textBG:SetPoint("TOPLEFT", button, 2, -2)
    textBG:SetPoint("BOTTOMRIGHT", button, -2, 2)
    textBG:SetColorTexture(0, 0, 0, 0.6)
    button.textBG = textBG

    local text = button:CreateFontString(nil, "BACKGROUND")
    text:SetFontObject(GGL.GetFont(12))
    text:SetTextColor(1, 1, 1)
    text:SetAllPoints(textBG)
    text:SetJustifyH("CENTER")
    text:SetJustifyV("MIDDLE")
    --text:SetWordWrap(false)
    button.text = text

    return button
end

local QUICK_INFO_FRAME_PROTO = {
    ["Clear"] = function(self)
        self:SetStatistics()
        self:SetPlayerTemplate()
        self:SetTopItems()
        self:SetItemQualityList()

        -- frames clear
        self:UpdateTextFromStatistics()
        self:ClearButtonList()
    end,

    ["ClearButtonList"] = function(self)
        local buttonFrame = self.frame.buttonFrame
        local count = #buttonFrame.buttons.shown
        for i = 1, count do
            buttonFrame.buttons.hidden[buttonFrame.buttons.shown[i]] = true
            buttonFrame.buttons.shown[i] = nil
        end
    end,

    ["AddNewButton"] = function(self, icon, text, numTextLines, isAtlasTexture)
        local buttonFrame = self.frame.buttonFrame
        numTextLines = numTextLines or 1

        local button = next(buttonFrame.buttons.hidden)
        if not button then
            button = CreateNewButton(buttonFrame.content)
        else
            buttonFrame.buttons.hidden[button] = nil
        end
        buttonFrame.buttons.shown[#buttonFrame.buttons.shown + 1] = button
        local buttonCount = #buttonFrame.buttons.shown

        button:ClearAllPoints()
        button:SetParent(buttonFrame.content)
        if buttonCount > 1 then
            button:SetPoint("TOP", buttonFrame.buttons.shown[buttonCount-1], "BOTTOM", 0, -2)
        else
            button:SetPoint("TOP", buttonFrame.content)
        end
        if isAtlasTexture then
            button.icon:SetAtlas(icon)
        else
            button.icon:SetTexture(icon)
        end
        button.text:SetText(text)
        button:Show()

        -- resize frame
        buttonFrame:ClearAllPoints()
        buttonFrame:SetParent(self.frame)
        if self.frame.oriantationX == "RIGHT" then
            buttonFrame:SetPoint("TOPLEFT", self.frame, "TOPRIGHT", 3, 0)
        else
            buttonFrame:SetPoint("TOPRIGHT", self.frame, "TOPLEFT", -3, 0)
        end
        buttonFrame:SetHeight((buttonCount * QIF_BUTTON_FRAME_SIZE) + ((buttonCount-1) * 2) + 20)
        buttonFrame:Show()
    end,

    --- Sets the statistics for the frame
    ["SetStatistics"] = function(self, statistics)
        self.statistics = statistics
        self:UpdateFromStatistics()
    end,

    --- Updates the gui with info from a statistics register
    -- Needs a player template registered
    ["UpdateFromStatistics"] = function(self, playerTemplate, statistics, addPlayerIfMissing)
        if playerTemplate then self:SetPlayerTemplate(playerTemplate) end
        if statistics then return self:SetStatistics(statistics) end
        if not self.statistics then return end
        if not self.playerTemplate then return end

        local player = self.statistics:GetPlayerTable(self.playerTemplate, nil, not addPlayerIfMissing)
        if not player then return end

        self:SetItemQualityList(player.itemQualityCount, player.numItems)
        self:SetTopItems(self.statistics:GetTopListForMergeID(self.playerTemplate))
        self:SetActivityPercent(self.statistics:GetActivityPercentForMergeID(self.playerTemplate))
        self:SetKillCount(self.statistics:GetKillCountForMergeID(self.playerTemplate))

        -- button list
        -- self:AddNewButton(236174, format( "Test:\n%d", 99999) )

        -- set raw gold
        self:UpdateTextFromStatistics()
    end,

    ["UpdateTextFromStatistics"] = function(self)
        local frame = QuickInfoFrame.frame

        if self.statistics then
            local statistics = self.statistics
            -- gold
            frame.iniBox.rawGold:SetText(GetMoneyTextShort(statistics:GetRawGoldForMergeID(self.playerTemplate)))
            frame.iniBox.itemValue:SetText(GetMoneyTextShort(statistics:GetItemValueGoldForMergeID(self.playerTemplate)))
            -- player
            local numEntered, numReset = statistics:GetInstanceResetDataForMergeID(self.playerTemplate)
            frame.iniBox.iniEntered:SetText(numEntered or 0)
            frame.iniBox.iniReset:SetText(numReset or 0)
            frame.iniBox.usedChars:SetText(statistics:GetNumCharsForMergeID(self.playerTemplate))
            -- island expedition box
            local fullCount, ieData = statistics:GetIslandExpeditionDataForMergeID(self.playerTemplate)
            if fullCount then
                frame.islandExpeditionBox:AddIntoFrame(true)
                for i = 1, 3 do
                    local data = ieData[i]
                    if i == 1 then
                        frame.islandExpeditionBox.lines[1].text:SetText(data and (data[1] or 0) or 0)
                        frame.islandExpeditionBox.lines[4].text:SetText(data and (data[2] or 0) or 0)
                    elseif i == 2 then
                        frame.islandExpeditionBox.lines[2].text:SetText(data and (data[1] or 0) or 0)
                        frame.islandExpeditionBox.lines[5].text:SetText(data and (data[2] or 0) or 0)
                    elseif i == 3 then
                        frame.islandExpeditionBox.lines[3].text:SetText(data and (data[1] or 0) or 0)
                        frame.islandExpeditionBox.lines[6].text:SetText(data and (data[2] or 0) or 0)
                    end
                end
            else
                frame.islandExpeditionBox:AddIntoFrame(false)
            end
        else
            -- gold
            frame.iniBox.rawGold:SetText(GetMoneyTextShort(0))
            frame.iniBox.itemValue:SetText(GetMoneyTextShort(0))
            -- player
            frame.iniBox.iniEntered:SetText(0)
            frame.iniBox.iniReset:SetText(0)
            frame.iniBox.usedChars:SetText(0)
            frame.iniBox.killCount:SetText(0)
            -- island expedition box
            frame.islandExpeditionBox:AddIntoFrame(false)
        end
        --GetNumCharsForMergeID
    end,

    --- Sets a player template for live updates
    -- @param       playerTemplate      <playerTemplate>
    ["SetPlayerTemplate"] = function(self, playerTemplate)
        if self.playerTemplate then
            self.playerTemplate:UnregisterCallbackEvent("PlayerTemplate_PlayerUpdate", self.CB_OnPlayerUpdate)
            self.playerTemplate = nil
        end
        if not playerTemplate then
            return
        end


        self.playerTemplate = playerTemplate
        playerTemplate:RegisterCallbackEvent("PlayerTemplate_PlayerUpdate", self.CB_OnPlayerUpdate)

        local bTagBox = QuickInfoFrame.frame.bTagBox

        -- set battleTag
        if playerTemplate.bnPlayer and playerTemplate.bnPlayer.battleTag then
            bTagBox.text:SetText(playerTemplate.bnPlayer.battleTag)
        else
            bTagBox.text:SetText("")
        end

        -- set name
        if playerTemplate.className then
            bTagBox.name:SetText(format(PLAYER_PLAYING_FORMAT_COLOR, GGL.CLASS_COLORS[GGL.LOCALIZED_CLASS_TO_NAME[playerTemplate.className]].colorStr, playerTemplate.characterName or UNKNOWN, playerTemplate.realmName or ""))
        else
            bTagBox.name:SetText(format(PLAYER_PLAYING_FORMAT, playerTemplate.characterName, playerTemplate.characterName))
        end

        -- set faction icon
        if playerTemplate.factionName and GGL.FACTION_ICON_ATLAS[playerTemplate.factionName] then
            bTagBox.factionIcon:SetAtlas(GGL.FACTION_ICON_ATLAS[playerTemplate.factionName])
            bTagBox.factionIcon:Show()
        else
            bTagBox.factionIcon:Hide()
        end

        -- set ggl version
        if playerTemplate.gglVersion then
            bTagBox.version:SetText(playerTemplate.gglVersion)
            bTagBox.version:Show()
        else
            bTagBox.version:Hide()
        end

        self.CB_OnPlayerUpdate(playerTemplate)
    end,

    --- Sets the top item
    -- @param       topItems        <table> {itemID, value}
    ["SetTopItems"] = function(self, topItems)
        local buttons = QuickInfoFrame.frame.topItemsBox.buttons

        for i = 1, #buttons do
            local button, item = buttons[i], topItems and topItems[i] or nil
            if topItems and item then
                local itemName, itemLink = GetItemInfo(item[1])

                button.icon:SetTexture(GetItemIcon(item[1]))
                button.itemName:SetFormattedText(ITEM_NUM_FORMAT, self.statistics:GetItemCountForMergeID(self.playerTemplate, item[1]), itemLink or "")
                button.itemValue:SetText(GetMoneyTextShort(item[2]))

                button.itemID = item[1]

                button:Show()
            else
                button.itemID = nil
                button:Hide()
            end
        end
    end,

    --- Sets item quality count
    -- @param       qualityList        <table> [qualityID] = {count, percent}
    ["SetItemQualityList"] = function(self, qualityList, numItems)
        local qFrames = QuickInfoFrame.frame.itemQualityBox.qFrames

        for i = 0, #qFrames do
            if qFrames[i] then
                if qualityList and qualityList[i] and qualityList[i] > 0 and numItems > 0 then
                    qFrames[i].textPercent:SetFormattedText(PERCENT_NUMBER_FORMAT, (100/(numItems))*qualityList[i])
                    qFrames[i].textCount:SetFormattedText(qualityList[i] or 0)
                else
                    qFrames[i].textPercent:SetFormattedText(PERCENT_NUMBER_FORMAT, 0)
                    qFrames[i].textCount:SetFormattedText(0)
                end
            end
        end
    end,

    --- Sets activity percent
    -- @param       percent         <number>
    ["SetActivityPercent"] = function(self, percent)
        local bar = QuickInfoFrame.frame.activityBar.bar

        percent = percent or 0
        percent = percent > 100 and 100 or percent

        if percent >= 75 then
            bar:SetBarColor(0, 1, 0)
        elseif percent >= 50 then
            bar:SetBarColor(1, 1, 0)
        elseif percent >= 25 then
            bar:SetBarColor(1, 0.533, 0)
        else
            bar:SetBarColor(1, 0, 0)
        end

        bar:SetValue(1, 100, percent, L["Activity: %s"], true)
    end,

    --- Sets kill count
    -- @param       count         <number>
    ["SetKillCount"] = function(self, count)
        self.frame.iniBox.killCount:SetText(count or 0)
    end,

    -- CB's
    ["CB_OnPlayerUpdate"] = function(playerTemplate)
        local bTagBox = QuickInfoFrame.frame.bTagBox

        -- set zone
        if playerTemplate.areaName then
            bTagBox.zone:SetText(playerTemplate.areaName)
        else
            bTagBox.zone:SetText(UNKNOWN)
        end
    end,
}

local function QuickInfoFrame_Item_OnEnter(self)
    if not self.itemID then return end
    local tooltip = _G.GameTooltip
    tooltip:ClearLines()
    tooltip:SetOwner(UIParent,"ANCHOR_CURSOR")
    tooltip:SetItemByID(self.itemID)
    tooltip:Show()
end

local function QuickInfoFrame_Item_OnLeave(self)
    _G.GameTooltip:Hide()
end

local function QuickInfoFrame_AddItemButton(parent)
    local height = 12
    local frame = CreateFrame("Frame", nil, parent)
    frame:SetScript("OnEnter", QuickInfoFrame_Item_OnEnter)
    frame:SetScript("OnLeave", QuickInfoFrame_Item_OnLeave)
    frame:SetHeight(height)

    local icon = frame:CreateTexture(nil, "BACKGROUND")
    icon:SetPoint("LEFT", frame, "LEFT", 0, 0)
    icon:SetWidth(height)
    icon:SetHeight(height)
    frame.icon = icon

    local itemName = frame:CreateFontString(nil, "BACKGROUND")
    itemName:SetPoint("LEFT", icon, "RIGHT", 1, 0)
    itemName:SetFontObject(GGL.GetFont(height))
    itemName:SetPoint("RIGHT", frame, 0, 0)
    itemName:SetJustifyH("LEFT")
    itemName:SetJustifyV("MIDDLE")
    itemName:SetWordWrap(false)
    itemName:SetHeight(height)
    frame.itemName = itemName

    local itemValue = frame:CreateFontString(nil, "BACKGROUND", "GameFontHighlightSmall")
    itemValue:SetAllPoints(itemName)
    itemValue:SetFontObject(GGL.GetFont(height))
    itemValue:SetJustifyH("RIGHT")
    itemValue:SetJustifyV("MIDDLE")
    itemValue:SetWordWrap(false)
    itemValue:SetHeight(height)
    frame.itemValue = itemValue

    return frame
end

local function QuickInfoFrame_UpdateFrameHeight()
    if not QuickInfoFrame then return end
    QuickInfoFrame.frame:SetHeight(QuickInfoFrame.frame.baseHeight + QuickInfoFrame.frame.modHeight)
end

local function CreateNewStatisticEntry(iniBox, entryHeight, descHeight, text)
    local entry = iniBox:CreateFontString(nil, "BORDER", nil, 2)
    entry:SetFontObject(GGL.GetFont(entryHeight))
    entry:SetHeight(entryHeight)
    entry:SetJustifyH("RIGHT")
    entry:SetJustifyV("MIDDLE")
    entry:SetText(999)

    local desc = iniBox:CreateFontString(nil, "BORDER", nil, 2)
    desc:SetFontObject(GGL.GetFont(descHeight))
    desc:SetAllPoints(entry)
    desc:SetJustifyH("LEFT")
    desc:SetJustifyV("MIDDLE")
    desc:SetTextColor(0.625, 0.625, 0.625)
    desc:SetText(text)

    return entry, desc
end

local function CreateQuickInfoFrame()
    if not QuickInfoFrame then
        local fullFrameHeight = 20

        local frame = GGL.GUI:CreateBoxFrame(UIParent)
        frame:SetPoint("CENTER", UIParent, 100, 0)
        frame:SetWidth(270)
        frame:SetHeight(350)
        frame:SetBackdropColor(0, 0, 0, 0.7)
        frame:SetBackdropBorderColor(0, 0, 0, 1)

        local content = CreateFrame("FRAME", nil, frame)
        content:SetPoint("TOPLEFT", frame, 10, -10)
        content:SetPoint("BOTTOMRIGHT", frame, -10, 10)
        frame.content = content

        -- #######
        -- button frame
        local buttonFrame = GGL.GUI:CreateBoxFrame(frame)
        buttonFrame:SetPoint("CENTER", frame, 100, 0)
        buttonFrame:SetWidth(QIF_BUTTON_FRAME_SIZE + 10)
        buttonFrame:SetHeight(QIF_BUTTON_FRAME_SIZE + 10)
        buttonFrame:SetBackdropColor(0, 0, 0, 0.7)
        buttonFrame:SetBackdropBorderColor(0, 0, 0, 1)
        buttonFrame:Hide()
        buttonFrame.buttons = {
            shown = {},
            hidden = {}
        }
        frame.buttonFrame = buttonFrame

        local buttonFrameContent = CreateFrame("FRAME", nil, buttonFrame)
        buttonFrameContent:SetPoint("TOPLEFT", buttonFrame, 10, -10)
        buttonFrameContent:SetPoint("BOTTOMRIGHT", buttonFrame, -10, 10)
        buttonFrame.content = buttonFrameContent
        -- end button frame
        -- #######


        -- #######
        -- battleTag
        local bTagBox = Templates.GetBoxWithHeader(content)
        bTagBox.text:SetText(L["Battle.Tag"])
        bTagBox:SetPoint("TOPLEFT", content, 0, -5)
        bTagBox:SetPoint("RIGHT", content, 0, 0)
        bTagBox:SetHeight(48)
        frame.bTagBox = bTagBox

        bTagBox.factionIcon = bTagBox.content:CreateTexture(nil, "BACKGROUND")
        bTagBox.factionIcon:SetPoint("RIGHT", bTagBox, -1, 0)
        bTagBox.factionIcon:SetWidth(bTagBox:GetHeight()-2)
        bTagBox.factionIcon:SetHeight(bTagBox:GetHeight()-2)
        bTagBox.factionIcon:SetAlpha(0.25)

        bTagBox.version = bTagBox.content:CreateFontString(nil, "BORDER", nil, 2)
        bTagBox.version:SetFontObject(GGL.GetFont(12))
        bTagBox.version:SetHeight(12)
        bTagBox.version:SetPoint("TOPRIGHT", bTagBox, -2, -2)
        bTagBox.version:SetPoint("LEFT", bTagBox, 0, 0)
        bTagBox.version:SetJustifyH("RIGHT")
        bTagBox.version:SetJustifyV("MIDDLE")
        bTagBox.version:SetAlpha(0.5)
        bTagBox.version:SetText(UNKNOWN)

        bTagBox.name = bTagBox.content:CreateFontString(nil, "BORDER", nil, 2)
        bTagBox.name:SetFontObject(GGL.GetFont(12))
        bTagBox.name:SetHeight(12)
        bTagBox.name:SetPoint("TOPLEFT", bTagBox.content, 0, 0)
        bTagBox.name:SetPoint("RIGHT", bTagBox.content, 0, 0)
        bTagBox.name:SetJustifyH("LEFT")
        bTagBox.name:SetJustifyV("MIDDLE")
        bTagBox.name:SetText(UNKNOWN)

        bTagBox.zoneText = bTagBox.content:CreateFontString(nil, "BORDER", nil, 2)
        bTagBox.zoneText:SetFontObject(GGL.GetFont(16))
        bTagBox.zoneText:SetHeight(16)
        bTagBox.zoneText:SetPoint("TOPLEFT", bTagBox.name, "BOTTOMLEFT", 0, -2)
        bTagBox.zoneText:SetJustifyH("LEFT")
        bTagBox.zoneText:SetJustifyV("MIDDLE")
        bTagBox.zoneText:SetText(L["Zone:"])
        bTagBox.zoneText:SetWidth(bTagBox.zoneText:GetUnboundedStringWidth()+2)

        bTagBox.zone = bTagBox.content:CreateFontString(nil, "BORDER", "GameFontNormal", 2)
        bTagBox.zone:SetFontObject(GGL.GetFont(16))
        bTagBox.zone:SetHeight(16)
        bTagBox.zone:SetPoint("LEFT", bTagBox.zoneText, "RIGHT", 0, 0)
        bTagBox.zone:SetPoint("RIGHT", bTagBox.content, 0, 0)
        bTagBox.zone:SetJustifyH("LEFT")
        bTagBox.zone:SetJustifyV("MIDDLE")
        bTagBox.zone:SetText(UNKNOWN)

        fullFrameHeight = fullFrameHeight + 5 + bTagBox:GetHeight()
        -- battleTag
        -- #######

        -- #######
        -- activity
        local activityBox = GGL.GUI:CreateBoxFrame(content)
        activityBox:SetHeight(22)
        activityBox:SetPoint("TOPLEFT", bTagBox, "BOTTOMLEFT", 0, -5)
        activityBox:SetPoint("RIGHT", content,  0, 0)
        frame.activityBox = activityBox

        local activityBar = Templates.GetProgressBar(content)
        activityBar:SetHeight(18)
        activityBar.bar:SetHeight(18)
        activityBar:SetPoint("TOPLEFT", activityBox, 2, -2)
        activityBar:SetPoint("RIGHT", activityBox, -2, 0)
        frame.activityBar = activityBar

        fullFrameHeight = fullFrameHeight + 27
        -- activity
        -- #######

        -- #######
        -- left > reset box
        local iniBoxFullWidth = content:GetWidth() * 0.66

        local iniBox = GGL.GUI:CreateBoxFrame(content)
        iniBox:SetHeight(66) -- 58
        iniBox:SetPoint("TOPLEFT", activityBox, "BOTTOMLEFT", 0, -5)
        iniBox:SetWidth(iniBoxFullWidth)
        frame.iniBox = iniBox

        iniBoxFullWidth = iniBoxFullWidth - 10

        local entryHeight, descHeight, gap = 14, 12, 5
        local entryCounter = 0

        iniBox.iniEntered, iniBox.iniEnteredDesc = CreateNewStatisticEntry(iniBox, entryHeight, descHeight, L["Instances:"])
        iniBox.iniEntered:SetPoint("TOPLEFT", iniBox, 5, -6)
        iniBox.iniEntered:SetPoint("RIGHT", iniBox, -5, 0)
        entryCounter = entryCounter + 1

        iniBox.iniReset, iniBox.iniResetDesc = CreateNewStatisticEntry(iniBox, entryHeight, descHeight, L["Instance Resets:"])
        iniBox.iniReset:SetPoint("TOPLEFT", iniBox.iniEntered, "BOTTOMLEFT", 0, -(gap))
        iniBox.iniReset:SetPoint("RIGHT", iniBox.iniEntered)
        entryCounter = entryCounter + 1

        iniBox.killCount, iniBox.killCountDesc = CreateNewStatisticEntry(iniBox, entryHeight, descHeight, L["Kills:"])
        iniBox.killCount:SetPoint("TOPLEFT", iniBox.iniReset, "BOTTOMLEFT", 0, -(gap))
        iniBox.killCount:SetPoint("RIGHT", iniBox.iniEntered)
        entryCounter = entryCounter + 1

        iniBox.usedChars, iniBox.usedCharsDesc = CreateNewStatisticEntry(iniBox, entryHeight, descHeight, L["Characters:"])
        iniBox.usedChars:SetPoint("TOPLEFT", iniBox.killCount, "BOTTOMLEFT", 0, -(gap))
        iniBox.usedChars:SetPoint("RIGHT", iniBox.iniEntered)
        entryCounter = entryCounter + 1

        iniBox.rawGold, iniBox.rawGoldDesc = CreateNewStatisticEntry(iniBox, entryHeight, descHeight, L["Raw gold:"])
        iniBox.rawGold:SetPoint("TOPLEFT", iniBox.usedChars, "BOTTOMLEFT", 0, -(gap))
        iniBox.rawGold:SetPoint("RIGHT", iniBox.iniEntered)
        entryCounter = entryCounter + 1

        iniBox.itemValue, iniBox.itemValueDesc = CreateNewStatisticEntry(iniBox, entryHeight, descHeight, L["Item value:"])
        iniBox.itemValue:SetPoint("TOPLEFT", iniBox.rawGold, "BOTTOMLEFT", 0, -(gap))
        iniBox.itemValue:SetPoint("RIGHT", iniBox.iniEntered)
        entryCounter = entryCounter + 1

        iniBox:SetHeight(12 + (entryCounter * entryHeight) + ((entryCounter-1) * gap))
        -- left > reset box end
        -- #######


        -- #######
        -- right > itemQuality box
        local itemQualityBox = GGL.GUI:CreateBoxFrame(content)
        itemQualityBox:SetPoint("TOPRIGHT", activityBox, "BOTTOMRIGHT", 0, -5)
        itemQualityBox:SetPoint("BOTTOMLEFT", iniBox, "BOTTOMRIGHT", 5, 0)
        frame.itemQualityBox = itemQualityBox

        itemQualityBox.qFrames = {}
        local qFrameWidth = itemQualityBox:GetWidth()
        local qFrameHeight = (iniBox:GetHeight() - 10) / (#GGL.ITEM_QUALITY_INFO + 1)
        local textSize = 11
        for quality = 0, #GGL.ITEM_QUALITY_INFO do
            local qData = GGL.ITEM_QUALITY_INFO[quality]

            local qFrame = CreateFrame("FRAME", nil, itemQualityBox)
            qFrame:SetPoint("LEFT", itemQualityBox)
            qFrame:SetPoint("RIGHT", itemQualityBox)
            qFrame:SetHeight(qFrameHeight)

            if quality == 0 then
                qFrame:SetPoint("TOPLEFT", itemQualityBox, 0, -5)
            else
                qFrame:SetPoint("TOPLEFT", itemQualityBox.qFrames[quality-1], "BOTTOMLEFT", 0, 0)
            end

            local icon = qFrame:CreateTexture(nil, "BACKGROUND")
            icon:SetPoint("CENTER", qFrame, "CENTER")
            icon:SetWidth(textSize + 2)
            icon:SetHeight(textSize + 2)
            icon:SetColorTexture(qData.color.r, qData.color.g, qData.color.b)
            qFrame.icon = icon

            local textCount = qFrame:CreateFontString(nil, "BACKGROUND")
            textCount:SetPoint("RIGHT", icon, "LEFT", -3, 0)
            textCount:SetFontObject(GGL.GetFont(textSize))
            textCount:SetHeight(textSize)
            textCount:SetPoint("LEFT", qFrame, 0, 0)
            textCount:SetJustifyH("RIGHT")
            textCount:SetJustifyV("MIDDLE")
            textCount:SetWordWrap(false)
            textCount:SetFormattedText(9999)
            qFrame.textCount = textCount

            local textPercent = qFrame:CreateFontString(nil, "BACKGROUND")
            textPercent:SetPoint("LEFT", icon, "RIGHT", 0, 0)
            textPercent:SetPoint("RIGHT", qFrame, -5, 0)
            textPercent:SetFontObject(GGL.GetFont(textSize))
            textPercent:SetHeight(textSize)
            textPercent:SetJustifyH("RIGHT")
            textPercent:SetJustifyV("MIDDLE")
            textPercent:SetWordWrap(false)
            textPercent:SetFormattedText(PERCENT_NUMBER_FORMAT, 100)
            qFrame.textPercent = textPercent

            itemQualityBox.qFrames[quality] = qFrame
        end
        -- right > itemQuality box
        -- #######

        fullFrameHeight = fullFrameHeight + 5 + iniBox:GetHeight()

        -- #######
        -- topItems box
        local topItemsBox = Templates.GetBoxWithHeader(content)
        topItemsBox.text:SetText(L["TopItems"])
        topItemsBox:SetPoint("TOP", iniBox, "BOTTOM", 0, -10)
        topItemsBox:SetPoint("LEFT", content, 0, 0)
        topItemsBox:SetPoint("RIGHT", content, 0, 0)
        topItemsBox:SetHeight(18 + (TOP_ITEMS_LIST_SIZE*12) + ((TOP_ITEMS_LIST_SIZE-1)*2))
        frame.topItemsBox = topItemsBox

        topItemsBox.buttons = {}
        for i = 1, TOP_ITEMS_LIST_SIZE do
            local button = QuickInfoFrame_AddItemButton(topItemsBox.content)

            if i == 1 then
                button:SetPoint("TOPLEFT", topItemsBox.content, 0, 0)
            else
                button:SetPoint("TOPLEFT", topItemsBox.buttons[i-1], "BOTTOMLEFT", 0, -2)
            end
            button:SetPoint("RIGHT", topItemsBox.content, 0, 0)

            topItemsBox.buttons[i] = button
        end

        fullFrameHeight = fullFrameHeight + 10 + topItemsBox:GetHeight()
        -- topItems box
        -- #######

        -- #######
        -- island expedition box
        local islandExpeditionBox = GGL.GUI:CreateBoxFrame(content)
        islandExpeditionBox:SetPoint("TOPLEFT", itemQualityBox, "BOTTOMLEFT", 0, -5)
        islandExpeditionBox:SetPoint("RIGHT", itemQualityBox, 0, 0)
        islandExpeditionBox:Hide()
        islandExpeditionBox.state = false
        islandExpeditionBox.AddIntoFrame = function(islandExpeditionBoxSelf, state)
            if state == islandExpeditionBox.state then return end
            islandExpeditionBox.state = state
            if state then
                frame.modHeight = frame.modHeight + islandExpeditionBox.height
                islandExpeditionBox:Show()
            else
                frame.modHeight = frame.modHeight - islandExpeditionBox.height
                islandExpeditionBox:Hide()
            end
            QuickInfoFrame_UpdateFrameHeight()
        end
        frame.islandExpeditionBox = islandExpeditionBox

        islandExpeditionBox.icon = islandExpeditionBox:CreateTexture(nil, "BACKGROUND")
        islandExpeditionBox.icon:SetPoint("LEFT", islandExpeditionBox, 2, 0)
        islandExpeditionBox.icon:SetWidth(40)
        islandExpeditionBox.icon:SetHeight(40)
        islandExpeditionBox.icon:SetAtlas("poi-islands-table")
        --islandExpeditionBox.factionIcon:SetAlpha(0.25)

        islandExpeditionBox.solo = islandExpeditionBox:CreateFontString(nil, "BACKGROUND")
        islandExpeditionBox.solo:SetPoint("TOPLEFT", islandExpeditionBox.icon, "TOPRIGHT", 2, 0)
        islandExpeditionBox.solo:SetFontObject(GGL.GetFont(12))
        islandExpeditionBox.solo:SetWidth(100)
        islandExpeditionBox.solo:SetHeight(20)
        islandExpeditionBox.solo:SetJustifyH("LEFT")
        islandExpeditionBox.solo:SetJustifyV("MIDDLE")
        islandExpeditionBox.solo:SetWordWrap(false)
        islandExpeditionBox.solo:SetText(L["Solo:"])

        islandExpeditionBox.group = islandExpeditionBox:CreateFontString(nil, "BACKGROUND")
        islandExpeditionBox.group:SetPoint("TOPLEFT", islandExpeditionBox.solo, "BOTTOMLEFT", 0, 0)
        islandExpeditionBox.group:SetFontObject(GGL.GetFont(12))
        islandExpeditionBox.group:SetWidth(100)
        islandExpeditionBox.group:SetHeight(20)
        islandExpeditionBox.group:SetJustifyH("LEFT")
        islandExpeditionBox.group:SetJustifyV("MIDDLE")
        islandExpeditionBox.group:SetWordWrap(false)
        islandExpeditionBox.group:SetText(L["Group:"])

        local ieDescTextWidth = math.max(islandExpeditionBox.solo:GetStringWidth(), islandExpeditionBox.group:GetStringWidth())
        islandExpeditionBox.solo:SetWidth(ieDescTextWidth)
        islandExpeditionBox.group:SetWidth(ieDescTextWidth)

        local fullLineWidth = islandExpeditionBox:GetWidth() - (islandExpeditionBox.icon:GetWidth() + ieDescTextWidth + 4 + 2)
        fullLineWidth = fullLineWidth / 3
        local ieFrameHeight = 20
        islandExpeditionBox.lines = {}
        for i = 1, 6 do
            local ieFrame = CreateFrame("FRAME", nil, islandExpeditionBox)
            ieFrame:SetWidth(fullLineWidth)
            ieFrame:SetHeight(ieFrameHeight)

            local icon = ieFrame:CreateTexture(nil, "BACKGROUND")
            icon:SetPoint("LEFT", ieFrame, 0, 0)
            icon:SetWidth(ieFrameHeight)
            icon:SetHeight(ieFrameHeight)
            ieFrame.icon = icon

            local text = ieFrame:CreateFontString(nil, "BACKGROUND")
            text:SetPoint("LEFT", icon, "RIGHT", 1, 0)
            text:SetFontObject(GGL.GetFont(12))
            text:SetHeight(12)
            text:SetPoint("RIGHT", ieFrame, 0, 0)
            text:SetJustifyH("LEFT")
            text:SetJustifyV("MIDDLE")
            text:SetWordWrap(false)
            text:SetText(123)
            ieFrame.text = text

            if i == 1 or i == 4 then
                icon:SetAtlas("islands-queue-difficultyselector-1")
                if i == 1 then
                    ieFrame:SetPoint("TOPLEFT", islandExpeditionBox.solo, "TOPRIGHT", 2, 0)
                else
                    ieFrame:SetPoint("TOPLEFT", islandExpeditionBox.lines[1], "BOTTOMLEFT", 0, 0)
                end
            elseif i == 2 or i == 5 then
                icon:SetAtlas("islands-queue-difficultyselector-2")
                ieFrame:SetPoint("LEFT", islandExpeditionBox.lines[i-1], "RIGHT", 0, 0)
            elseif i == 3 or i == 6 then
                icon:SetAtlas("islands-queue-difficultyselector-3")
                ieFrame:SetPoint("LEFT", islandExpeditionBox.lines[i-1], "RIGHT", 0, 0)
            end

            islandExpeditionBox.lines[i] = ieFrame
        end

        islandExpeditionBox.height = 5 + (ieFrameHeight * 2) + 2
        islandExpeditionBox:SetHeight(islandExpeditionBox.height)

        -- island expedition box
        -- #######

        frame:SetHeight(fullFrameHeight)
        frame.baseHeight = fullFrameHeight
        frame.modHeight = 0

        local window = {
            frame = frame,
        }
        setmetatable(window, {__index=QUICK_INFO_FRAME_PROTO})

        QuickInfoFrame = window
    end
    return QuickInfoFrame
end

--- Shows the quickInfoFrame
-- @param       parent          <frame>
-- @param       attachedSide    <string> <default>"LEFT", "RIGHT"
-- @param       ...             <points> :SetPoint(...)
function Statistics:ShowQuickInfoFrame(parent, attachedSide, ...)
    local frame = CreateQuickInfoFrame()
    frame:Clear()
    frame.frame:ClearAllPoints()
    frame.frame:SetParent(parent)
    frame.frame:SetPoint(...)
    frame.frame:Show()
    frame.oriantationX = attachedSide or "LEFT"
    frame.oriantationY = "BOTTOM"
    return frame
end

--- Shows the quickInfoFrame as Tooltip
-- @param       parent          <frame>
-- @param       point           <string> <default> "BOTTOM_X"
function Statistics:ShowQuickInfoFrameAsTooltip(parent, point, pointX, pointY)
    local frame = CreateQuickInfoFrame()
    frame:Clear()
    frame.oriantationX, frame.oriantationY = GGL.GUI:SetToolTipPoint_Frame(frame.frame, parent, point or "BOTTOM_X", pointX, pointY)
    frame.frame:Show()
    return frame
end

-- ##################
-- GUI
-- ##################
local FrameTypeRegister = {}
local WIDTH_V, HEIGHT_V = 600, 500
local WIDTH_M, HEIGHT_M = 500, 500

local function BaseFrameDropDown_OnValueChanged(self, event, value)
    self = self.obj or self
    local frame = self.frame

    -- create frame if not existing
    if not frame.contentGroupRegister[value] then
        local contentGroup = AceGUI:Create("SimpleGroup")
        contentGroup:ClearAllPoints()
        contentGroup.frame:SetParent(frame.content)
        contentGroup:SetLayout("Flow")
        contentGroup.frame:SetPoint("TOPLEFT", frame.dropdown.frame, "BOTTOMLEFT", 0, -3)
        contentGroup.frame:SetPoint("BOTTOMRIGHT", frame.buttonGroup.frame, "BOTTOMLEFT", -3, 0)
        contentGroup.frame.widget = contentGroup
        frame.contentGroupRegister[value] = contentGroup
    end

    -- hide old content
    if self.shownContent then
        -- call OnHide register
        if self.onHideList and self.uNameList[self.shownID] and self.onHideList[self.uNameList[self.shownID]] then
            self.onHideList[self.uNameList[self.shownID]](self, self.private[self.shownID], self.shownContent.frame)
        end
        self.shownContent.frame:Hide()
    end

    -- set new frame
    self.shownContent = frame.contentGroupRegister[value]
    self.shownID = value
    self.shownContent.frame:Show()
    if self.executeRegister[value] then
        self.executeRegister[value](self, self.private[value], self.shownContent.frame)
        self.executeRegister[value] = nil
    end
    if self.onShowList and self.uNameList[value] and self.onShowList[self.uNameList[value]] then
        self.onShowList[self.uNameList[value]](self, self.private[value], self.shownContent.frame)
    end
end

local function CreateBaseFrame(frameType)
    local window = AceGUI:Create("Window")
    window:SetTitle("GoldGoblinLoot - "..frameType.name)

    local frame = window.frame
    local content = window.content

    -- some frame settings
    frame:SetMinResize(WIDTH_V,HEIGHT_V)
    frame:SetMaxResize(WIDTH_V+WIDTH_M,HEIGHT_V+HEIGHT_M)
    frame:SetClampedToScreen(true)
    frame:SetToplevel(false)
    frame:SetFrameStrata("MEDIUM")
    --window settings
    window:SetWidth(WIDTH_V)
    window:SetHeight(HEIGHT_V)
    -- remove resizer lines
    window.line1:Hide()
    window.line2:Hide()

    local buttonGroup = AceGUI:Create("SimpleGroup")
    buttonGroup:ClearAllPoints()
    buttonGroup.frame:SetParent(content)
    buttonGroup:SetLayout("List")
    buttonGroup:SetWidth(150)
    buttonGroup.frame:SetPoint("TOPRIGHT", content, 0, 0)
    buttonGroup.frame:SetPoint("BOTTOM", content, 0, 0)
    buttonGroup.obj = frameType
    window.buttonGroup = buttonGroup

    local dropdown = AceGUI:Create("Dropdown")
    dropdown:ClearAllPoints()
    dropdown.frame:SetParent(content)
    dropdown:SetCallback("OnValueChanged", BaseFrameDropDown_OnValueChanged)
    dropdown.frame:SetPoint("TOPLEFT", content, 0, 0)
    dropdown.frame:SetPoint("RIGHT", buttonGroup.frame, "LEFT", -3, 0)
    dropdown.frame:Show()
    dropdown:SetDisabled(true)
    dropdown.obj = frameType
    window.dropdown = dropdown

    window.contentGroupRegister = {}

    GGL.GUI:AddFrameControl("StatisticsFrame-"..frameType.uName, frame, true, false)

    return window
end

local function NpcExportButton_OnClick(self, event, button)
    if GoldGoblinLootDataDB then
        Templates.ShowExportFrame(GoldGoblinLootDataDB["npc_names"], "csv")
    end
end

local BASE_FRAME_PROTO = {
    -- #########
    -- register base

    --- Sets a list for the top dropdown
    -- @param       list        <table> { uName, displayName, guiFillFunc(self, content) }
    ["SetDropDownList"] = function(self, list)
        if not self.frame then
            self.loadDropDownList = list
            return
        end
        local dropdown = self.frame.dropdown
        if not list then
            dropdown:SetDisabled(true)
            dropdown:SetList()
            self.executeRegister = nil
            self.uNameList = nil
            self.private = nil
        else
            dropdown:SetDisabled(false)

            self.uNameList = {}
            self.private = {}
            local fillList, sortList, executeRegister = {}, {}, {}
            for i = 1, #list do
                local entry = list[i]
                sortList[i] = i
                fillList[i] = entry[2]
                if type(entry[3]) == "string" then
                    -- use the function of another entry
                    for j = 1, #list do
                        if list[j][1] == entry[3] then
                            executeRegister[i] = list[j][3]
                            break
                        end
                    end

                else
                    executeRegister[i] = entry[3]
                end
                self.uNameList[entry[1]] = i
                self.uNameList[i] = entry[1]
                self.private[i] = {
                    uName = entry[1],
                    name = entry[2],
                }
            end
            dropdown:SetList(fillList, sortList)
            self.executeRegister = executeRegister
            dropdown:SetValue(1)
            BaseFrameDropDown_OnValueChanged(dropdown, "OnValueChanged", 1)
        end
    end,

    --- Sets onShowlist
    -- @param       list        <table> ["uName"] = function(self, content)
    ["SetOnShowList"] = function(self, list)
        self.onShowList = list
    end,

    --- Sets onHideList
    -- @param       list        <table> ["uName"] = function(self, content)
    ["SetOnHideList"] = function(self, list)
        self.onHideList = list
    end,

    --- Add a new button in the right buttonList
    -- @param       text        <string>
    -- @param       onClickCB   <function>
    -- @param       ...         <strings> Tooltip lines
    ["AddButton"] = function(self, text, onClickCB, ...)
        for i = 1, #self.frame.buttonGroup.children do
            if self.frame.buttonGroup.children.text == text then
                return
            end
        end
        local button = AceGUI:Create("Button")
        button:SetText(text)
        button:SetFullWidth(true)
        button:SetCallback("OnClick", onClickCB)
        GGL.GUI:AddToolTipForAceGUIWidget(button, text, ...)
        self.frame.buttonGroup:AddChild(button)
    end,

    --- Add a npc export button for all npcs
    ["AddNpcExportButton"] = function(self)
        self:AddButton(L["Export NPC Names"], NpcExportButton_OnClick)
    end,

    --- Clear all buttons from the list
    ["ClearButtons"] = function(self)
        self.frame.buttonGroup:ReleaseChildren()
    end,

    --- Opens a specific fram by its unique name
    -- @param       uName       <string> unique name
    ["OpenFrame"] = function(self, uName)
        if not self.uNameList or not self.uNameList[uName] then return end
        self.frame.dropdown:SetValue(self.uNameList[uName])
        BaseFrameDropDown_OnValueChanged(self, "OnValueChanged", self.uNameList[uName])
    end,
    -- register base
    -- #########

    --- Shows the main frame
    ["Show"] = function(self)
        self.frame.frame:Show()
    end,
}

--- Register a new frame type
-- @param       uName       <string> unique name
-- @param       name        <string> display name
-- @returns methods from: BASE_FRAME_PROTO
function Statistics:RegisterFrameType(uName, name)
    if FrameTypeRegister[uName] then return end

    local frameType = {
        uName = uName,
        name = name,
    }

    for k,v in pairs(BASE_FRAME_PROTO) do
        frameType[k] = v
    end

    GGL:AddCallbackHandler(frameType)

    FrameTypeRegister[uName] = frameType

    return frameType, Templates
end

--- Loads the frame of a registered type
-- @param       uName           <string> unique name registered with RegisterFrameType
-- @param       statistics      <core:statistics> statistics table to load
function Statistics:Load(uName, statistics)
    if not uName or not FrameTypeRegister[uName] or not statistics then return end
    local loadType = FrameTypeRegister[uName]

    -- create frame if its not exist
    if not loadType.frame then
        loadType.frame = CreateBaseFrame(loadType)
    end

    -- loaded dropdown list
    if loadType.loadDropDownList then
        loadType:SetDropDownList(loadType.loadDropDownList)
        loadType.loadDropDownList = nil
    end

    -- set statistics
    loadType.statistics = statistics

    -- show frame
    loadType.frame:Show()

    return loadType
end


function _G.GGL_TEST()
    Statistics:Load("baseFrame", {})
end