local AddonName, GGL = ...
local Templates = GGL.Templates
local Sender = GGL.Sender
local Session = Templates.Session
local SyncStatic = Templates.Sync

local Challenge, ChallengeHelper = Templates:Register("Challenge", "Session", "Sync")

-- lua
local sub, match = string.sub, string.match

-- wow
local C_TimerAfter = C_Timer.After

-- GGL
local GetPlayerFromSync = Templates.Sync.GetPlayerFromSync

-- ##################
-- local
-- ##################
local SessionSave = {}

local INVITE_STATES = GGL.INVITE_KEY_LIST

local PLAYER_STATUS_UPDATE_SEND = "%d\002%d\002%s\002%s\002%s\002%s\002%s\002%s\002%s" -- rank, attendees, curGold, upGold, upName, downGold, downName, firstGold, firstName
local PLAYER_STATUS_UPDATE_GET = "^(%d*)\002(%d*)\002(.*)\002(.*)\002(.*)\002(.*)\002(.*)\002(.*)\002(.*)"

-- set timers only one time
local TimerRegister = {}
local function Set_C_TimerAfter(after, method)
    if not TimerRegister[method] then
        TimerRegister[method] = true
        C_TimerAfter(after, method)
    end
end

-- ##################
-- static
-- ##################
local Static = {}
Templates.Challenge = Static

function Static:GetSession(sessionDB, sessionID)
    sessionDB = sessionDB or GGL.sessionDb.Sessions
    if sessionID and SessionSave[sessionID] then
        return SessionSave[sessionID], false
    else
        return Templates:New("Challenge", sessionDB, sessionID), true
    end
end

-- ##################
-- template
-- ##################
Challenge.sessionFormat = "C%d"
Challenge.sessionType = "C"

-- #################
-- Sender
-- #################

local function Sender_OnPlayerStatusUpdate(self, prefix, sender, message, isBNMessage)
    local session = SessionSave[sub(prefix, 1, -2)]
    if not session then return end

    -- get playerTemplate
    sender = SyncStatic.GetPlayerFromSync(sender)
    if not sender or session.inviteTemplate:GetHost() ~= sender then return end

    local rank, attendees, curGold, upGold, upName, downGold, downName, firstGold, firstName = match(message, PLAYER_STATUS_UPDATE_GET)
    if not rank then return end

    session:UpdateOwnData(rank, attendees, tonumber(curGold), tonumber(upGold), upName, tonumber(downGold), downName, tonumber(firstGold), firstName)
end

-- Instance reset
local function Sender_OnPlayerInstanceResetUpdate(self, prefix, sender, message, isBNMessage)
    local session = SessionSave[sub(prefix, 1, -3)]
    if not session then return end

    -- get playerTemplate
    sender = SyncStatic.GetPlayerFromSync(sender)
    if not sender then return end

    local pass, data = GGL:Deserialize(message)
    if not pass or not data then return end
    session:UpdateResetData(sender, data)
end

-- Island Expedition
local function Sender_OnIslandExpeditionFinish(self, prefix, sender, message, isBNMessage)
    local session = SessionSave[sub(prefix, 1, -3)]
    if not session then return end

    -- get playerTemplate
    sender = SyncStatic.GetPlayerFromSync(sender)
    if not sender then return end

    local pass, data = GGL:Deserialize(message)
    if not pass or not data then return end

    session:IslandExpedition_FullUpdate(data, sender)
end

-- #################
-- Callbacks
-- #################
local function CB_HostStartsSession(self, session, duration, endTime, settingsString)
    session:SetData()
    session:SetSettingsString(settingsString)
    if not endTime or endTime == 0 then
        -- duration for endless is the startTime send from Host
        session.db.timeElapsed = GGL.GetTime() - duration
        session:SetTimerEndless()
    else
        session:SetTimerData(endTime - duration, endTime)
    end
    session:Start()
end

local function CB_HostEndsSession(self, session)
    session:Close()
end

local function CB_CFrameOnNewItem(self, itemID, itemString, zoneID, sourceID, ammount, copperValue, mainSourceFlag, subSourceFlag, sourcePlayer)
    if self.challengeFrame then
        self.challengeFrame:SetPlayerGold(sourcePlayer, copperValue * ammount, false)
    end
    if self.topItemsFrame then
        self.topItemsFrame:AddItem(sourcePlayer, itemID, self:LoggedItemUseVariations(itemID, itemString) and itemString or nil, ammount, copperValue, zoneID, sourceID, mainSourceFlag, subSourceFlag)
    end
end

local function CB_CFrameOnAddMoney(self, copper, sourcePlayer)
    if not self.challengeFrame then return end
    --self.challengeFrame:SetGoldCount(copper, true)
    self.challengeFrame:SetPlayerGold(sourcePlayer, copper, false)
end

local function CB_CFrameOnNewSyncLoot(self, sourcePlayer, lootTime, zoneIniID, difficultyID, itemID, itemLog, ammount, sourceID, mainSourceFlag, subSourceFlag)
    self.pauseLog = true
    if itemID == 0 then
        self:AddMoney(ammount, nil, sourceID, mainSourceFlag, subSourceFlag, nil, zoneIniID, difficultyID, lootTime, nil, sourcePlayer)
    else
        self:AddNewItem(itemID, itemLog, ammount, nil, sourceID, mainSourceFlag, subSourceFlag, nil, zoneIniID, difficultyID, lootTime, nil, sourcePlayer)
    end
    self.pauseLog = nil
end

local function CB_CFrameOnPlayerAccepted(self, player)
    if self.challengeFrame then
        self.challengeFrame:AddPlayer(player)
        self.challengeFrame:SetPlayerInviteState(player, INVITE_STATES.ACCEPTED)
    end
    if self.topItemsFrame then
        self.topItemsFrame:SetPlayerInviteState(player, INVITE_STATES.ACCEPTED)
    end
end

local function CB_CFrameOnPlayerDeclined(self, player)
    if self.challengeFrame then
        self.challengeFrame:SetPlayerInviteState(player, INVITE_STATES.DECLINED)
    end
    if self.topItemsFrame then
        self.topItemsFrame:SetPlayerInviteState(player, INVITE_STATES.DECLINED)
    end
end

local function CB_CFrameOnPlayerInvited(self, player)
    if self.challengeFrame then
        self.challengeFrame:SetPlayerInviteState(player, INVITE_STATES.INVITED)
    end
    if self.topItemsFrame then
        self.topItemsFrame:SetPlayerInviteState(player, INVITE_STATES.INVITED)
    end
end

local function CB_CFrameOnPlayerMergeIDUpdate(self)
    if self.challengeFrame then
        self.challengeFrame:UpdateMergeList()
    end
    if self.topItemsFrame then
        self.topItemsFrame:UpdateMergeList()
    end
    self:IslandExpedition_UpdateMergeData()
end

local function CB_SetDataRefresh(self)
    if self.topItemsFrame then
        self:GUI_RemoveFromTopItemsFrame()
        self:GUI_AddIntoTopItemsFrame(true)
    end
    if self.challengeFrame then
        self:GUI_RemoveFromChallengeFrame()
        self:GUI_AddIntoChallengeFrame(true)
    end

    self:RebuildFromLog()
    CB_CFrameOnPlayerMergeIDUpdate(self)
end

-- #################
-- ChallengeFrame
-- #################
function Challenge:GUI_RemoveFromChallengeFrame()
    if not self.challengeFrame then return end
    local frame = self.challengeFrame

    frame.timer:FinishTimer()

    frame.runningSession = nil

    frame:SetUpdateInterval(nil)

    -- remove from minimap button
    GGL.GUI.Challenge:RemoveFromMiniMapButton()

    -- one last update
    frame:UpdateList()

    -- freeze the list
    frame:UnregisterPlayerUpdates()

    -- show the export button
    frame:ShowExportButton(true)

    self.challengeFrame = nil
end

function Challenge:GUI_AddIntoChallengeFrame(add)
    if not self.db or self.db.isClosed or not self:PlayerIsHost() then return end
    if add then
        local frame = GGL.GUI.Challenge:CreateFrame()

        -- start timer if its a running session or just set timer
        local db = self.db
        if self.db.endlessTime  then
            if db.isRunning  then
                frame.timer:SetEndless(db.timeElapsed)
                frame.timer:StartTimer()
                frame:Show()
            end
        elseif db.isRunning then
            frame.timer:SetTime(db.startTime, db.endTime)
            frame.timer:IsCounter(false)
            frame.timer:StartTimer()
        elseif db.durationTime then
            frame.timer:SetDuration(db.durationTime)
        end

        -- set data
        if self.challengeFrame ~= frame then
            self.challengeFrame = frame

            -- clear old frame
            frame:Clear()
            self.rebuildSync = nil

            -- set session for challenge frame
            frame:SetSession(self)

            -- first fill with lootLog data
            for playerSyncID in pairs(self.db.lootLog) do
                if playerSyncID ~= self.selfDbKey then
                    local playerTemplateLootLog = Templates.Sync.GetPlayerFromSync(playerSyncID, true)
                    frame:AddPlayer(playerTemplateLootLog)
                end
            end
            -- set chars
            frame:AddPlayer(self.selfPlayer)
            local playerList = self.playerList
            for player, inviteState in self:IterInvitedPlayers() do
                if (inviteState == INVITE_STATES.ACCEPTED) or (self.db and self.db.lootLog and self.db.lootLog[self:GetPlayerSyncID(player)]) then
                    frame:AddPlayer(player)
                end
            end

            -- update
            frame:SetUpdateInterval(15)

            -- add into minimap button
            GGL.GUI.Challenge:AddIntoMiniMapButton()

            -- show the export button
            frame:ShowExportButton(false)

            -- update
            frame:UpdateMergeList()
        end
    else
        self:GUI_RemoveFromChallengeFrame()
    end
end

-- #################
-- SmallChallengeFrame
-- #################
function Challenge:GUI_RemoveFromSmallChallengeFrame()
    if not self.smallChallengeFrame then return end
    local frame = self.smallChallengeFrame

    frame.timer:FinishTimer()

    frame.runningSession = nil

    -- remove from minimap button
    GGL.GUI.Challenge:RemoveSmallFromMiniMapButton()

    self.smallChallengeFrame = nil
end

function Challenge:GUI_AddIntoSmallChallengeFrame(add)
    if not self.db or self.db.isClosed then return end

    local frame = GGL.GUI.Challenge:ShowSmallFrame()

    -- start timer if its a running session or just set timer
    local db = self.db
    if self.db.endlessTime  then
        if db.isRunning  then
            frame.timer:SetEndless(db.timeElapsed)
            frame.timer:StartTimer()
            frame:Show()
        end
    elseif db.isRunning then
        frame.timer:SetTime(db.startTime, db.endTime)
        frame.timer:IsCounter(false)
        frame.timer:StartTimer()
    elseif db.durationTime then
        frame.timer:SetDuration(db.durationTime)
    end

    -- clear all
    frame:Clear()

    -- add into small frame
    GGL.GUI.Challenge:AddSmallIntoMiniMapButton()

    self.smallChallengeFrame = frame
end

-- #################
-- TopItems
-- #################
function Challenge:GUI_RemoveFromTopItemsFrame()
    if not self.topItemsFrame then return end
    local frame = self.topItemsFrame

    -- Unregister all player events
    frame:UnregisterAllPlayerEvents()

    -- remove from minimap button
    GGL.GUI.TopItems:RemoveFromMiniMapButton()

    self.topItemsFrame = nil
end

function Challenge:GUI_AddIntoTopItemsFrame(add)
    if not self.db or self.db.isClosed then return end

    local frame = GGL.GUI.TopItems:Show()

    -- clear all
    frame:Clear()

    -- set session
    frame:SetSession(self)

    -- update
    frame:UpdateMergeList()

    -- add into small frame
    GGL.GUI.TopItems:AddIntoMiniMapButton()

    self.topItemsFrame = frame
end

-- #################
-- Challenge
-- #################
function Challenge:Constructor(sessionDB, sessionID)
    if sessionID and SessionSave[sessionID] then
        return SessionSave[sessionID]
    end
    -- base
    self:CallBaseFunction("Session", "Constructor", sessionDB, sessionID)
    self:CallBaseFunction("Sync", "Constructor")

    -- sync settings with players
    self:SyncSettingsWithPlayers(true)

    self.inviteTemplate:RegisterCallbackEvent("SessionTemplate_Invite_HostStartsSession", CB_HostStartsSession)
    self.inviteTemplate:RegisterCallbackEvent("SessionTemplate_Invite_HostEndsSession", CB_HostEndsSession)

    SessionSave[self.db.sessionID] = self
end

function Challenge:Start()
    -- callback
    self:RegisterCallbackEvent("SessionTemplate_AddItem", CB_CFrameOnNewItem)
    self:RegisterCallbackEvent("SessionTemplate_AddMoney", CB_CFrameOnAddMoney)
    -- always show main and challenge frame
    self:AddIntoMainFrame(true)
    self:GUI_AddIntoTopItemsFrame(true)
    if self:PlayerIsHost() then
        self:GUI_AddIntoChallengeFrame(true)
    end

    self:CallBaseFunction("Session", "Start")
    self:CallBaseFunction("Sync", "Start")
    if not self.db.isRunning then return end

    if self:PlayerIsHost() then
        self:RegisterCallbackEvent("SessionTemplate_Sync_NewSyncLoot", CB_CFrameOnNewSyncLoot)
        self:RegisterCallbackEvent("SessionTemplate_Sync_PlayerAccepted", CB_CFrameOnPlayerAccepted)
        self:RegisterCallbackEvent("SessionTemplate_Sync_PlayerDeclined" , CB_CFrameOnPlayerDeclined)
        self:RegisterCallbackEvent("SessionTemplate_Sync_PlayerInvited" , CB_CFrameOnPlayerInvited)
        self:RegisterCallbackEvent("SessionTemplate_Sync_PlayerMergeIDUpdate" , CB_CFrameOnPlayerMergeIDUpdate)

        self:GUI_AddIntoChallengeFrame(true)
        GGL.GUI.Challenge:Show()

        if not self.__playerStatusUpdate then
            self.__playerStatusUpdate = function()
                if not self.db.isRunning or not self.challengeFrame then
                    if TimerRegister and self.__playerStatusUpdate then
                        TimerRegister[self.__playerStatusUpdate] = nil
                    end
                    self.__playerStatusUpdate = nil
                    return
                end

                local challengeChildren = self.challengeFrame.charContainer.children
                local firstName, firstGold
                if challengeChildren and challengeChildren[1] then
                    firstName = challengeChildren[1]:GetPlayer().displayName or ""
                    firstGold = tostring(challengeChildren[1]:GetMoney() or 0)
                end
                for i = 1, #challengeChildren do
                    local childPrev = challengeChildren[i-1]
                    local child = challengeChildren[i]
                    local childNext = challengeChildren[i+1]
                    if child:GetPlayer() == self.selfPlayer then
                        self:UpdateOwnData(
                            child:GetRank(),
                            #challengeChildren,
                            child:GetMoney(),
                            childPrev and childPrev:GetMoney() or 0,
                            childPrev and childPrev:GetPlayer().displayName or "",
                            childNext and childNext:GetMoney() or 0,
                            childNext and childNext:GetPlayer().displayName or "",
                            tonumber(firstGold),
                            firstName
                        )
                    else
                        if self.inviteTemplate:GetPlayerInviteState(child:GetPlayer()) == INVITE_STATES.ACCEPTED then
                            -- rank, attendees, curGold, upGold, upName, downGold, downName
                            self.cSender.PlayerStatusUpdate:SendMessage(child:GetPlayer(),
                                format(PLAYER_STATUS_UPDATE_SEND,
                                    child:GetRank(),
                                    #challengeChildren,
                                    tostring(child:GetMoney()),
                                    childPrev and tostring(childPrev:GetMoney()) or "0",
                                    childPrev and childPrev:GetPlayer().displayName or "",
                                    childNext and tostring(childNext:GetMoney()) or "0",
                                    childNext and childNext:GetPlayer().displayName or "",
                                    firstGold,
                                    firstName
                                )
                            )
                        end
                    end
                end

                C_TimerAfter((#challengeChildren*0.1) + 15, self.__playerStatusUpdate)
            end
            Set_C_TimerAfter(16, self.__playerStatusUpdate)
        end
    end

    -- show small frame
    self:GUI_AddIntoSmallChallengeFrame(true)

    -- SetDataRefresh
    self:RegisterCallbackEvent("SessionTemplate_SetDataRefresh", CB_SetDataRefresh)

    -- Instance resets
    if not self.challenge_cbOnInstancesResetUpdate then
        self.challenge_cbOnInstancesResetUpdate = function(e, data)
            self:InstanceReset_SendDataToHost()
        end
    end
    GGL:RegisterCallbackEvent("GGL_OnInstancesResetUpdate", self.challenge_cbOnInstancesResetUpdate)

    -- island expedition counter
    if not self.challenge_cbOnIslandExpeditionFinish then
        self.challenge_cbOnIslandExpeditionFinish = function(e, mapID, difficultyID, isInParty, numPartyMember)
            self:IslandExpedition_OnFinish(mapID, difficultyID, isInParty, numPartyMember)
        end
    end
    GGL:RegisterCallbackEvent("GGL_OnIslandExpeditionFinish", self.challenge_cbOnIslandExpeditionFinish)

    -- sender
    self.cSender = self.cSender or {}

    if not self.cSender.PlayerStatusUpdate or not self.cSender.PlayerStatusUpdate.Unregister then
        self.cSender.PlayerStatusUpdate = GGL.Sender:RegisterReceiver(self.db.sessionID.."u", Sender_OnPlayerStatusUpdate)
    end
    if not self.cSender.InstanceResets or not self.cSender.InstanceResets.Unregister then
        self.cSender.InstanceResets = GGL.Sender:RegisterReceiver(self.db.sessionID.."ir", Sender_OnPlayerInstanceResetUpdate)
    end
    if not self.cSender.IslandExpeditionFinish or not self.cSender.IslandExpeditionFinish.Unregister then
        self.cSender.IslandExpeditionFinish = GGL.Sender:RegisterReceiver(self.db.sessionID.."ie", Sender_OnIslandExpeditionFinish)
    end

    -- update
    if self.topItemsFrame then
        self.topItemsFrame:UpdateMergeList()
    end

    -- send reset data
    self:InstanceReset_SendDataToHost()
    self:IslandExpedition_SendOwnDataToHost()

    return self.db.startTime, self.db.endTime, self.db.durationTime
end

function Challenge:SetData(mainSet, setRefresh)
    self:CallBaseFunction("Session", "SetData", mainSet, setRefresh)

    if self.challengeFrame then
        self:GUI_AddIntoChallengeFrame(true)
    end
end

--- Updates data for the small frame
-- @param       rank        <number> current rank
-- @param       attendees   <number> attendees in this session
-- @param       curGold     <number> (copper) current player money value
-- @param       upGold      <number> (copper) current player money for the rank over the player
-- @param       upName      <string> name of the player for the rank over the player
-- @param       downGold    <number> (copper) current player money for the rank under the player
-- @param       downName    <string> name of the player for the rank under the player
-- @param       firstGold    <number> (copper) first player money
-- @param       downName    <string> first player name
function Challenge:UpdateOwnData(rank, attendees, curGold, upGold, upName, downGold, downName, firstGold, firstName)
    if self.smallChallengeFrame then
        local smallFrame = self.smallChallengeFrame

        smallFrame:SetFirstPlayerData(firstGold, firstName, attendees)
        smallFrame:SetUpPlayerData(upGold, upName, rank-1)
        smallFrame:SetPlayerData(curGold, self.selfPlayer.displayName, rank)
        smallFrame:SetDownPlayerData(downGold, downName, rank+1)
    end
end

-- #################
-- Extra
-- #################

-- ## Instance reset
function Challenge:InstanceReset_SendDataToHost(data)
    if self:PlayerIsHost() then return end

    data = data or self:UpdateOwnResetData(true)

    self.cSender.InstanceResets:SendMessage(self:GetHostPlayer(), GGL:Serialize(data))
end

function Challenge:InstanceReset_GetData()
    return self.db.instanceResetData
end

function Challenge:InstanceReset_GetDataForPlayer(playerTemplate)
    if not playerTemplate then return end
    return self.db.instanceResetData[self:GetPlayerMergeID(playerTemplate)]
end
-- ## Instance reset end

-- ## Island Expedition
function Challenge:IslandExpedition_OnFinish(mapID, difficultyID, isInParty, numPartyMember, playerTemplate)
    local playerMergeID = self:GetPlayerMergeID(playerTemplate or self.selfPlayer)
    if not playerMergeID then return end

    if not self.db.islandExpeditionFinishData then self.db.islandExpeditionFinishData = {} end
    if not self.db.islandExpeditionFinishData[playerMergeID] then self.db.islandExpeditionFinishData[playerMergeID] = {} end

    self.db.islandExpeditionFinishData[playerMergeID][ #self.db.islandExpeditionFinishData[playerMergeID] + 1 ] = {mapID, difficultyID, isInParty, numPartyMember}

    if self:PlayerIsHost() then return end -- no need to send data

    self.cSender.IslandExpeditionFinish:SendMessage(self:GetHostPlayer(), GGL:Serialize(self.db.islandExpeditionFinishData[playerMergeID]))
end

function Challenge:IslandExpedition_UpdateMergeData()
    if not self.db.islandExpeditionFinishData then return end
    local islandExpeditionData = self.db.islandExpeditionFinishData

    for syncID in pairs(self.db.lootLog) do
        local playerTemplate = GetPlayerFromSync(syncID, true)
        if playerTemplate and playerTemplate.mergeID and islandExpeditionData[syncID] then
            local tmp = islandExpeditionData[syncID]
            islandExpeditionData[syncID] = nil
            self:IslandExpedition_FullUpdate(tmp, playerTemplate)
        end
    end
end

function Challenge:IslandExpedition_FullUpdate(newData, playerTemplate)
    playerTemplate = playerTemplate or self.selfPlayer

    local playerMergeID = self:GetPlayerMergeID(playerTemplate)
    if not playerMergeID then return end

    if not self.db.islandExpeditionFinishData then
        self.db.islandExpeditionFinishData = {}
    end

    self.db.islandExpeditionFinishData[playerMergeID] = newData
end

function Challenge:IslandExpedition_SendOwnDataToHost()
    if self:PlayerIsHost() or not self.db.islandExpeditionFinishData then return end
    local playerMergeID = self:GetPlayerMergeID(self.selfPlayer)
    if not playerMergeID or not self.db.islandExpeditionFinishData[playerMergeID] then return end

    self.cSender.IslandExpeditionFinish:SendMessage(self:GetHostPlayer(), GGL:Serialize(self.db.islandExpeditionFinishData[playerMergeID]))
end
-- ## Island Expedition end

-- #################
-- End
-- #################
local function UnregisterAll(self)
    self:GUI_RemoveFromChallengeFrame()
    self:GUI_RemoveFromSmallChallengeFrame()

    -- callback
    self:UnregisterCallbackEvent("SessionTemplate_AddItem", CB_CFrameOnNewItem)
    self:UnregisterCallbackEvent("SessionTemplate_AddMoney", CB_CFrameOnAddMoney)
    self:UnregisterCallbackEvent("SessionTemplate_Sync_NewSyncLoot", CB_CFrameOnNewSyncLoot)
    self:UnregisterCallbackEvent("SessionTemplate_Sync_PlayerAccepted", CB_CFrameOnPlayerAccepted)
    self:UnregisterCallbackEvent("SessionTemplate_Sync_PlayerDeclined" , CB_CFrameOnPlayerDeclined)
    self:UnregisterCallbackEvent("SessionTemplate_Sync_PlayerInvited" , CB_CFrameOnPlayerInvited)
    self:UnregisterCallbackEvent("SessionTemplate_Sync_PlayerMergeIDUpdate" , CB_CFrameOnPlayerMergeIDUpdate)
    GGL:UnregisterCallbackEvent("GGL_OnInstancesResetUpdate", self.challenge_cbOnInstancesResetUpdate)
    GGL:UnregisterCallbackEvent("GGL_OnIslandExpeditionFinish", self.challenge_cbOnIslandExpeditionFinish)

    if self.cSender then
        for senderName, sender in pairs(self.cSender) do
            if sender.Unregister then
                sender:Unregister()
            end
        end
    end

    SessionSave[self.db.sessionID] = nil
end

function Challenge:Unregister()
    self:CallBaseFunction("Session", "Unregister")
    self:CallBaseFunction("Sync", "Unregister")

    UnregisterAll(self)
end

function Challenge:Stop()
    self:CallBaseFunction("Session", "Stop")
    self:CallBaseFunction("Sync", "Stop")

    UnregisterAll(self)

    -- keep open
    self.inviteTemplate:RegisterCallbackEvent("SessionTemplate_Invite_HostStartsSession", CB_HostStartsSession)
    self.inviteTemplate:RegisterCallbackEvent("SessionTemplate_Invite_HostEndsSession", CB_HostEndsSession)
    SessionSave[self.db.sessionID] = self
end

--- Close the session
function Challenge:Close()
    -- send one last status update
    if self.__playerStatusUpdate then
        self.__playerStatusUpdate()
    end
    self:CallBaseFunction("Session", "Close")
    self:CallBaseFunction("Sync", "Close")

    UnregisterAll(self)
end